import email
import email.policy
import io
import json
import os
import requests
import sqlalchemy as sa
import sqlalchemy_utils as sau
import subprocess
import time
import urllib.parse
from dispatchsrht.tasks.gitlab.common import GitLabAuthorization
from dispatchsrht.tasks.gitlab.common import gitlabloginrequired
from dispatchsrht.tasks import TaskDef
from dispatchsrht.types import Task
from email.mime.text import MIMEText
from email.utils import parseaddr, getaddresses, formatdate, make_msgid
from flask import abort, Blueprint, redirect, render_template, request, url_for, session
from markupsafe import Markup
from srht.api import ensure_webhooks, encrypt_request_authorization, get_results
from srht.config import cfg, cfgb, cfgkeys, get_origin
from srht.crypto import verify_request_signature
from srht.email import start_smtp
from srht.database import Base, db
from srht.flask import icon, csrf_bypass
from srht.oauth import current_user, loginrequired
from srht.validation import Validation
from uuid import UUID, uuid4

_lists_sr_ht = get_origin("lists.sr.ht", default=None)
_repo_cache = cfg("dispatch.sr.ht::gitlab", "repo-cache", default=None)
_gitlab_enabled = cfgb("dispatch.sr.ht::gitlab", "enabled", default=False)

if _gitlab_enabled:
    from gitlab import Gitlab

# XXX: This should really not live in Python code, we should set up a celery
# worker or a Go daemon or something to deal with this asyncronously
class FileLock:
    def __init__(self, path):
        self.path = path

    def __enter__(self):
        while True:
            try:
                self.file = open(self.path, mode="x")
                break
            except Exception as ex:
                time.sleep(1)

    def __exit__(self, type, value, traceback):
        self.file.close()
        os.remove(self.path)

class PatchToGitLabMR(TaskDef):
    name = "patch_to_gitlab_mr"
    enabled = bool(_gitlab_enabled and _lists_sr_ht and _repo_cache)

    def description():
        return (icon("envelope-o") + Markup(" lists.sr.ht patches ") +
            icon("caret-right") + " " +
            icon("gitlab") + Markup(" GitLab merge requests"))

    class _PatchToGitLabMRRecord(Base):
        __tablename__ = "patch_to_gitlab_mr"
        id = sa.Column(sau.UUIDType, primary_key=True)
        created = sa.Column(sa.DateTime, nullable=False)
        updated = sa.Column(sa.DateTime, nullable=False)
        user_id = sa.Column(sa.Integer,
                sa.ForeignKey("user.id", ondelete="CASCADE"))
        user = sa.orm.relationship("User")
        task_id = sa.Column(sa.Integer,
                sa.ForeignKey("task.id", ondelete="CASCADE"))
        task = sa.orm.relationship("Task")
        repo_name = sa.Column(sa.Unicode, nullable=False)
        repo_id = sa.Column(sa.Integer, nullable=False)
        gitlab_webhook_id = sa.Column(sa.Integer, nullable=False)
        lists_webhook_id = sa.Column(sa.Integer, nullable=False)
        list_id = sa.Column(sa.Integer, nullable=False)
        list_name = sa.Column(sa.Unicode, nullable=False)
        upstream = sa.Column(sa.Unicode, nullable=False)

    blueprint = Blueprint("patch_to_gitlab_mr",
            __name__, template_folder="patch_to_gitlab_mr")

    def edit_GET(task):
        record = PatchToGitLabMR._PatchToGitLabMRRecord.query.filter(
            PatchToGitLabMR._PatchToGitLabMRRecord.task_id == task.id
        ).one_or_none()
        if not record:
            abort(404)
        return render_template("gitlab/edit.html", task=task, record=record)

    def edit_POST(task):
        return redirect(url_for("html.edit_task", task_id=task.id))

    @blueprint.route("/configure")
    @loginrequired
    def configure():
        canonical_upstream = cfg("dispatch.sr.ht::gitlab", "canonical-upstream")
        upstreams = [k for k in cfgkeys("dispatch.sr.ht::gitlab") if k not in [
            "enabled", "canonical-upstream", "repo-cache", canonical_upstream,
        ]]
        return render_template("gitlab/select-instance.html",
                canonical_upstream=canonical_upstream, upstreams=upstreams,
                instance_name=lambda inst: cfg("dispatch.sr.ht::gitlab",
                    inst).split(":")[0], details="""
                    This task will create a merge request for the selected
                    GitLab repository when a patch is received on a particular
                    lists.sr.ht mailing list.
                """)

    @blueprint.route("/configure/<upstream>/select-repo")
    @gitlabloginrequired
    def configure_repo_GET(gitlab, upstream):
        repos = gitlab.projects.list(membership=True)
        repos = sorted(repos, key=lambda r: r.attributes["name_with_namespace"])
        existing = PatchToGitLabMR._PatchToGitLabMRRecord.query.filter(
                PatchToGitLabMR._PatchToGitLabMRRecord.user_id == current_user.id,
                PatchToGitLabMR._PatchToGitLabMRRecord.upstream == upstream).all()
        existing = [e.repo_id for e in existing]
        return render_template("gitlab/select-repo.html",
                repos=repos, existing=existing, details="""
                    This task will create a merge request for the selected
                    GitLab repository when a patch is received on a particular
                    lists.sr.ht mailing list.
                """)

    @blueprint.route("/configure/<upstream>/select-repo", methods=["POST"])
    @gitlabloginrequired
    def configure_repo_POST(gitlab, upstream):
        valid = Validation(request)
        repo_id = valid.require("repo_id")
        if not valid.ok:
            abort(400)
        session["repo_id"] = int(repo_id)
        return redirect(url_for("patch_to_gitlab_mr.configure_list_GET",
            upstream=upstream))

    @blueprint.route("/configure/<upstream>/select-list")
    @gitlabloginrequired
    def configure_list_GET(gitlab, upstream):
        repo_id = session.get("repo_id")
        if repo_id is None:
            abort(400)
        lists = get_results(f"{_lists_sr_ht}/api/lists", current_user)
        existing = PatchToGitLabMR._PatchToGitLabMRRecord.query.filter(
                PatchToGitLabMR._PatchToGitLabMRRecord.user_id == current_user.id,
                PatchToGitLabMR._PatchToGitLabMRRecord.upstream == upstream).all()
        existing = [e.list_id for e in existing]
        return render_template("gitlab/select-list.html",
                lists=lists, existing=existing, details="""
                    This task will create a merge request for the selected
                    GitLab repository when a patch is received on a particular
                    lists.sr.ht mailing list.
                """, _lists_sr_ht=_lists_sr_ht)

    @blueprint.route("/configure/<upstream>/select-list", methods=["POSt"])
    @gitlabloginrequired
    def configure_list_POST(gitlab, upstream):
        valid = Validation(request)
        list_id = valid.require("list_id")
        if not valid.ok:
            abort(400)
        list_id = int(list_id)

        list_name = valid.require("list_name")
        if not valid.ok:
            abort(400)

        repo_id = session.get("repo_id")
        if repo_id is None:
            abort(400)

        project = gitlab.projects.get(repo_id)
        if not project:
            abort(400)

        r = requests.get(f"{_lists_sr_ht}/api/lists/{list_name}",
                headers=encrypt_request_authorization(current_user))
        if r.status_code != 200:
            raise Exception(r.json())
        mailing_list = r.json()

        task = Task()
        task.name = "{}::patch_to_gitlab_mr".format(
                project.attributes["name_with_namespace"])
        task.user_id = current_user.id
        task._taskdef = "patch_to_gitlab_mr"
        db.session.add(task)
        db.session.flush()

        record = PatchToGitLabMR._PatchToGitLabMRRecord()
        record.id = uuid4()
        record.user_id = current_user.id
        record.task_id = task.id
        record.gitlab_webhook_id = -1
        record.repo_name = project.attributes['name_with_namespace']
        record.repo_id = project.id
        record.upstream = upstream
        record.list_id = list_id
        record.list_name = list_name
        record.lists_webhook_id = -1
        db.session.add(record)
        db.session.flush()

        _root = get_origin("dispatch.sr.ht", external=True)
        hook = project.hooks.create({
            "url": _root + url_for("patch_to_gitlab_mr._gitlab_webhook",
                record_id=record.id),
            "merge_requests_events": 1,
            "note_events": 1,
        })
        record.gitlab_webhook_id = hook.id

        _root = get_origin("dispatch.sr.ht")
        config = {
            _root + url_for("patch_to_gitlab_mr._list_webhook",
                    record_id=record.id):
                ["post:received", "patchset:received"],
        }
        url = f"{_lists_sr_ht}/api/user/{current_user.canonical_name}/lists/{list_name}/webhooks"
        ensure_webhooks(current_user, url, config)

        db.session.commit()
        return redirect(url_for("html.edit_task", task_id=task.id))

    @csrf_bypass
    @blueprint.route("/webhook/<record_id>/list", methods=["POST"])
    def _list_webhook(record_id):
        record_id = UUID(record_id)
        record = PatchToGitLabMR._PatchToGitLabMRRecord.query.filter(
                PatchToGitLabMR._PatchToGitLabMRRecord.id == record_id
            ).one_or_none()
        if not record:
            return "Unknown hook " + str(record_id), 404
        auth = GitLabAuthorization.query.filter(
                GitLabAuthorization.user_id == record.user_id,
                GitLabAuthorization.upstream == record.upstream,
            ).first()
        if not auth:
            return "Invalid authorization for this hook"

        event = request.headers.get("X-Webhook-Event")
        payload = verify_request_signature(request)
        payload = json.loads(payload.decode('utf-8'))

        if event == "patchset:received":
            return PatchToGitLabMR._patchset_received(payload, record, auth)
        elif event == "post:received":
            return PatchToGitLabMR._email_received(payload, record, auth)
        else:
            return "Unknown event, but thanks anyway"

    def _patchset_received(payload, record, auth):
        gitlab = Gitlab(f"https://{record.upstream}",
                oauth_token=auth.oauth_token)
        project = gitlab.projects.get(record.repo_id)
        git_url = urllib.parse.urlparse(project.http_url_to_repo)
        git_url = git_url._replace(netloc=f"oauth2:{auth.oauth_token}@{git_url.netloc}")
        git_url = urllib.parse.urlunparse(git_url)

        patch_id = payload["id"]
        _lists_sr_ht = get_origin("lists.sr.ht")
        patch_url = f"{_lists_sr_ht}/{record.user.canonical_name}/{record.list_name}/patches/{patch_id}"
        patches = list(get_results(f"{_lists_sr_ht}/api/user/{record.user.canonical_name}/lists/{record.list_name}/patchsets/{patch_id}/patches", record.user))

        cover = None
        if payload["cover_letter"] != None:
            r = requests.get(f"{_lists_sr_ht}/emails/{payload['cover_letter']['id']}",
                headers=encrypt_request_authorization(record.user))
            cover = r.json()

        with FileLock(os.path.join(_repo_cache, f"lock.{record.id}")):
            repo_path = os.path.join(_repo_cache, f"repo.{record.id}")
            if not os.path.exists(repo_path):
                subprocess.run(["git", "clone", git_url, repo_path], check=True)
            subprocess.run(["git", "-C", repo_path, "fetch", "origin"], check=True)
            branch_name = f"patches/{patch_id}"
            subprocess.run(["git", "-C", repo_path, "checkout",
                "-b", branch_name, "origin/master"])
            subprocess.run(["git", "-C", repo_path, "am", "--abort"],
                    stderr=subprocess.DEVNULL)
            r = requests.get(patch_url + "/mbox", stream=True)
            try:
                subprocess.run(["git", "-C", repo_path, "am"],
                        input=r.content, check=True)
            except subprocess.CalledProcessError:
                patches = list(get_results(f"{_lists_sr_ht}/api/user/{record.user.canonical_name}/lists/{record.list_name}/patchsets/{patch_id}/patches", record.user))
                msg = email.message_from_string(patches[0]["envelope"], policy=email.policy.SMTP)
                # XXX: Finish me
                reply = MIMEText("""Hi!

We received your patch, but we were unable to apply it to the upstream
GitLab repository. Did you rebase?
""")
                reply["To"] = msg["From"]
                reply["From"] = cfg("mail", "smtp-from")
                reply["Cc"] = msg["To"]
                reply["In-Reply-To"] = msg["Message-ID"]
                reply["Subject"] = "Re: " + msg.get("Subject")
                reply["Date"] = formatdate()
                reply["Message-ID"] = make_msgid()
                print(reply.as_string())

                smtp = start_smtp()
                tos = reply.get_all('To', [])
                ccs = reply.get_all('Cc', [])
                recipients = set([a[1] for a in getaddresses(tos + ccs)])

                _root = get_origin("dispatch.sr.ht", external=True)
                domain = urllib.parse.urlparse(_root).netloc
                smtp.send_message(reply, f"no-reply@{domain}", recipients)
                smtp.quit()
                return
            subprocess.run(["git", "-C", repo_path, "push", "origin", branch_name])

            if cover:
                msg = email.message_from_string(
                    cover["envelope"], policy=email.policy.SMTP)
                for part in msg.walk():
                    if part.is_multipart():
                        continue
                    content_type = part.get_content_type()
                    [charset] = part.get_charsets("utf-8")
                    if content_type == 'text/plain':
                        body = part.get_payload(decode=True).decode(charset, errors="replace")
                        break
                details = body
            elif len(patches) == 1:
                out = subprocess.run(["git", "-C", repo_path, "log", "-1", "--format=%b"],
                        check=True, capture_output=True)
                details = out.stdout.decode().strip()
            else:
                details = None

            _lists_sr_ht = get_origin("lists.sr.ht", external=True)
            patch_url = f"{_lists_sr_ht}/{record.user.canonical_name}/{record.list_name}/patches/{patch_id}"
            mr = project.mergerequests.create({
                "source_branch": branch_name,
                "target_branch": "master",
                "title": payload["subject"],
                "description":
                (
                    f"Patchset from {payload['submitter']} via [{record.list_name} #{patch_id}]({patch_url}):\n\n" +
                    ((f"```\n" + details + "\n" + f"```") if details else "")
                ),
                "remove_source_branch": 1,
            })

            subprocess.run(["git", "-C", repo_path,
                "config", f"dispatch.mr-{mr.id}", str(patch_id)], check=True)

        return "Thanks!"

    def _email_received(payload, record, auth):
        msg = email.message_from_string(
            payload["envelope"], policy=email.policy.SMTP)
        message_id = msg["Message-ID"]
        if ".gitlab." in message_id:
            return "Thanks!"
        in_reply_to = msg["In-Reply-To"]
        if not in_reply_to or ".gitlab." not in in_reply_to:
            return "Thanks!"
        for part in msg.walk():
            if part.is_multipart():
                continue
            content_type = part.get_content_type()
            [charset] = part.get_charsets("utf-8")
            if content_type == 'text/plain':
                body = part.get_payload(decode=True).decode(charset, errors="replace")
                break
        ident = in_reply_to.split("@")[0]
        merge_id = int(ident.split(".")[-2])
        discussion_id = ident.split(".")[-1]
        print(ident, merge_id)

        gitlab = Gitlab(f"https://{record.upstream}",
                oauth_token=auth.oauth_token)
        project = gitlab.projects.get(record.repo_id)
        merge = project.mergerequests.get(merge_id)
        discussion = merge.discussions.get(discussion_id)
        discussion.notes.create({
            "body": (f"{msg['From']} replied via email:\n\n" + f"```\n{body}\n```\n\n")
        })
        return "Thanks!"

    @csrf_bypass
    @blueprint.route("/webhook/<record_id>/gitlab", methods=["POST"])
    def _gitlab_webhook(record_id):
        event = request.headers.get("X-Gitlab-Event")
        payload = json.loads(request.data.decode('utf-8'))
        record_id = UUID(record_id)
        record = PatchToGitLabMR._PatchToGitLabMRRecord.query.filter(
                PatchToGitLabMR._PatchToGitLabMRRecord.id == record_id
            ).one_or_none()
        if not record:
            return "Unknown hook " + str(record_id), 404

        if event == "Note Hook":
            return PatchToGitLabMR._note_event(record, payload)
        elif event == "Merge Request Hook":
            return PatchToGitLabMR._mr_event(record, payload)
        else:
            return "Unknown event, but thanks anyway"

    def _note_event(record, payload):
        if not "merge_request" in payload:
            return "Unknown event, but thanks anyway"

        mr = payload["merge_request"]
        with FileLock(os.path.join(_repo_cache, f"lock.{record.id}")):
            repo_path = os.path.join(_repo_cache, f"repo.{record.id}")
            try:
                out = subprocess.run(["git", "-C", repo_path,
                    "config", f"dispatch.mr-{mr['id']}"],
                    check=True, capture_output=True)
            except subprocess.CalledProcessError:
                return "Unable to identify branch on our end"
            patch_id = int(out.stdout.decode().strip())

        r = requests.get(f"{_lists_sr_ht}/api/user/{record.user.canonical_name}/lists/{record.list_name}/patchsets/{patch_id}",
                headers=encrypt_request_authorization(record.user))
        patch = r.json()
        patches = list(get_results(f"{_lists_sr_ht}/api/user/{record.user.canonical_name}/lists/{record.list_name}/patchsets/{patch_id}/patches", record.user))
        obj = payload["object_attributes"]

        pos = obj["position"]
        if pos:
            for p in patches:
                msg = email.message_from_string(p["envelope"],
                        policy=email.policy.SMTP)
                for part in msg.walk():
                    if part.is_multipart():
                        continue
                    content_type = part.get_content_type()
                    [charset] = part.get_charsets("utf-8")
                    if content_type == 'text/plain':
                        body = part.get_payload(decode=True).decode(charset, errors="replace")
                        break

                context = []
                def yieldlines(body):
                    for line in body.splitlines():
                        context.append(line)
                        if len(context) > 3:
                            context.pop(0)
                        yield line

                CHANGE_NONE = 0
                CHANGE_ADD = 1
                CHANGE_REM = 2

                # TODO: Deal with moved files
                # XXX: A better way of going about this might be to just ask GitLab
                # for the relevant parts of the diff, but that seems somewhat
                # unappetizing to me at the moment.
                def yieldchanges(diff):
                    in_diff = False
                    path = oldno = newno = None
                    for line in diff:
                        if not in_diff:
                            in_diff = line.startswith("diff --git")
                            continue

                        if line.startswith("+++ "):
                            path = line[len("+++ "):]
                            if path.startswith("b/"):
                                path = path[2:]
                        elif line.startswith("--- "):
                            pass # TODO: deal with removed files
                        elif line.startswith("@@ "):
                            old, new = line.split(" ")[1:3]
                            old = old.split(",")[0]
                            new = new.split(",")[0]
                            oldno = int(old[1:]) # -123
                            newno = int(new[1:]) # +123
                        elif line.startswith("+"):
                            yield path, CHANGE_ADD, oldno, newno
                            newno += 1
                        elif line.startswith("-"):
                            yield path, CHANGE_REM, oldno, newno
                            oldno += 1
                        elif line.startswith(" "):
                            oldno += 1
                            newno += 1

                for change in yieldchanges(yieldlines(body)):
                    path, kind, old, new = change
                    if path != pos["old_path"]:
                        continue
                    if (   (kind == CHANGE_ADD and new == pos["new_line"] and pos["old_line"] == None)
                        or (kind == CHANGE_REM and new == None and pos["old_line"] == old)
                        or (kind == CHANGE_NONE and old == pos["old_line"] and new == pos["new_line"])):
                        context = "\n".join("> " + l for l in context)
                        break
                if isinstance(context, str):
                    break
        else:
            msg = email.message_from_string(patches[0]["envelope"],
                    policy=email.policy.SMTP)
            context = None

        msgfrom = parseaddr(msg["From"])
        if isinstance(context, str):
            body = f"On {msg['Date']}, {msgfrom[0]} wrote:\n{context}\n\n"
            body += obj['description']
        else:
            body = obj['description']
        body += f"\n\n-- \nvia {obj['url']}\n"

        user = payload['user']
        _root = get_origin("dispatch.sr.ht", external=True)
        domain = urllib.parse.urlparse(_root).netloc
        if 'name' in user:
            user = f"{user['name']} <no-reply@{domain}>"
        else:
            user = f"@{user['username']} <no-reply@{domain}>"

        message_id = make_msgid(f"gitlab.{mr['iid']}.{obj['discussion_id']}")
        in_reply_to = msg["Message-ID"]
        with FileLock(os.path.join(_repo_cache, f"lock.{record.id}")):
            repo_path = os.path.join(_repo_cache, f"repo.{record.id}")
            out = subprocess.run(["git", "-C", repo_path,
                "config", f"dispatch.discussion-{obj['discussion_id']}"],
                capture_output=True)
            if out.returncode != 0:
                ids = []
            else:
                ids = out.stdout.decode().strip().split(",")
                in_reply_to = ids[-1]
            ids.append(message_id)
            subprocess.run(["git", "-C", repo_path,
                "config", f"dispatch.discussion-{obj['discussion_id']}",
                ",".join(ids)], check=True)

        reply = MIMEText(body)
        reply["To"] = msg["From"]
        reply["From"] = user
        reply["Cc"] = msg["To"]
        reply["In-Reply-To"] = in_reply_to
        reply["Subject"] = "Re: " + (msg.get("Subject") or patch["subject"])
        reply["Date"] = formatdate()
        reply["Message-ID"] = message_id
        print(reply.as_string())

        smtp = start_smtp()
        tos = reply.get_all('To', [])
        ccs = reply.get_all('Cc', [])
        recipients = set([a[1] for a in getaddresses(tos + ccs)])
        smtp.send_message(reply, f"no-reply@{domain}", recipients)
        smtp.quit()
        return "Thanks!"

    def _mr_event(record, payload):
        obj = payload["object_attributes"]
        with FileLock(os.path.join(_repo_cache, f"lock.{record.id}")):
            repo_path = os.path.join(_repo_cache, f"repo.{record.id}")
            try:
                out = subprocess.run(["git", "-C", repo_path,
                    "config", f"dispatch.mr-{obj['id']}"],
                    check=True, capture_output=True)
            except:
                return "Ignoring merge request"
            patch_id = int(out.stdout.decode().strip())

        state = None
        if obj["action"] == "approved":
            state = "approved"
        elif obj["action"] == "unapproved":
            state = "proposed"
        elif obj["action"] == "merge":
            state = "applied"
        elif obj["action"] == "close":
            state = "rejected"
        elif obj["action"] == "reopen":
            state = "proposed"

        if state == None:
            return "Nothing to do here, but thanks anyway"

        r = requests.put(f"{_lists_sr_ht}/api/user/{record.user.canonical_name}/lists/{record.list_name}/patchsets/{patch_id}",
                headers=encrypt_request_authorization(record.user),
                json={
                    "status": state,
                })
        assert r.status_code == 200
        return "Thanks!"
